@@ -12,6 +12,7 @@ import psutil
 from aiodocker.docker import Docker, DockerContainer
 from aiodocker.exceptions import DockerError
 
+from ai.backend.agent.types import MountInfo
 from ai.backend.common.logging import BraceStyleAdapter
 from ai.backend.common.types import (
     DeviceId,
@@ -305,6 +306,16 @@ class CPUPlugin(AbstractComputePlugin):
                 )
         return attached_devices
 
+    async def get_docker_networks(
+        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
+    ) -> List[str]:
+        return []
+
+    async def generate_mounts(
+        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
+    ) -> List[MountInfo]:
+        return []
+
 
 class MemoryDevice(AbstractComputeDevice):
     pass
@@ -335,17 +346,17 @@ class MemoryPlugin(AbstractComputePlugin):
         pass
 
     async def list_devices(self) -> Collection[MemoryDevice]:
-        # TODO: support NUMA?
         memory_size = psutil.virtual_memory().total
         overcommit_factor = int(os.environ.get("BACKEND_MEM_OVERCOMMIT_FACTOR", "1"))
         return [
             MemoryDevice(
                 device_id=DeviceId("root"),
+                device_name=self.key,
                 hw_location="root",
-                numa_node=0,
+                numa_node=0,  # the kernel setting will do the job.
                 memory_size=overcommit_factor * memory_size,
                 processing_units=0,
-            )
+            ),
         ]
 
     async def available_slots(self) -> Mapping[SlotName, Decimal]:
@@ -613,3 +624,13 @@ class MemoryPlugin(AbstractComputePlugin):
                     }
                 )
         return attached_devices
+
+    async def get_docker_networks(
+        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
+    ) -> List[str]:
+        return []
+
+    async def generate_mounts(
+        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
+    ) -> List[MountInfo]:
+        return []
