@@ -20,6 +20,7 @@ from ai.backend.common.docker import ImageRef
 from ai.backend.common.logging import BraceStyleAdapter
 from ai.backend.common.types import KernelId
 from ai.backend.common.utils import current_loop
+from ai.backend.plugin.entrypoint import scan_entrypoints
 
 from ..kernel import AbstractCodeRunner, AbstractKernel
 from ..resources import KernelResourceSpec
@@ -202,6 +203,15 @@ class KubernetesKernel(AbstractKernel):
         result = await self.runner.feed_service_apps()
         return result
 
+    async def check_duplicate_commit(self, kernel_id, subdir):
+        log.error("Committing in Kubernetes is not supported yet.")
+        raise NotImplementedError
+
+    async def commit(self, kernel_id, subdir, filename):
+        # TODO: Implement container commit on Kubernetes kernel.
+        log.error("Committing in Kubernetes is not supported yet.")
+        raise NotImplementedError
+
     async def accept_file(self, filename: str, filedata: bytes):
         loop = current_loop()
         work_dir = self.agent_config["container"]["scratch-root"] / str(self.kernel_id) / "work"
@@ -251,6 +261,11 @@ class KubernetesKernel(AbstractKernel):
 
         return None
 
+    async def download_single(self, filepath: str):
+        # TODO: Implement download single file operations with pure Kubernetes API
+        log.error("download_single() in the k8s backend is not supported yet.")
+        raise NotImplementedError
+
     async def list_files(self, container_path: str):
         # TODO: Implement file operations with pure Kubernetes API
         await kube_config.load_kube_config()
@@ -479,8 +494,8 @@ async def prepare_krunner_env(local_config: Mapping[str, Any]) -> Mapping[str, S
 
     all_distros = []
     entry_prefix = "backendai_krunner_v10"
-    for entrypoint in pkg_resources.iter_entry_points(entry_prefix):
-        log.debug("loading krunner pkg: {}", entrypoint.module_name)
+    for entrypoint in scan_entrypoints(entry_prefix):
+        log.debug("loading krunner pkg: {}", entrypoint.module)
         plugin = entrypoint.load()
         await plugin.init({})  # currently does nothing
         provided_versions = (
