@@ -3,6 +3,7 @@ from __future__ import annotations
 import asyncio
 import functools
 import importlib
+import json
 import logging
 import logging.config
 import os
@@ -15,6 +16,7 @@ from ipaddress import ip_network
 from pathlib import Path
 from pprint import pformat, pprint
 from typing import (
+    TYPE_CHECKING,
     Any,
     AsyncGenerator,
     Callable,
@@ -23,6 +25,7 @@ from typing import (
     Dict,
     Literal,
     Mapping,
+    Optional,
     Sequence,
     Set,
     Tuple,
@@ -43,10 +46,14 @@ from setproctitle import setproctitle
 from trafaret.dataerror import DataError as TrafaretDataError
 
 from ai.backend.common import config, identity, msgpack, utils
+from ai.backend.common.bgtask import BackgroundTaskManager
 from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
+from ai.backend.common.events import EventProducer, KernelLifecycleEventReason
 from ai.backend.common.logging import BraceStyleAdapter, Logger
 from ai.backend.common.types import (
     ClusterInfo,
+    CommitStatus,
+    EtcdRedisConfig,
     HardwareMetadata,
     HostPortPair,
     KernelCreationConfig,
@@ -57,7 +64,6 @@ from ai.backend.common.types import (
 from ai.backend.common.utils import current_loop
 
 from . import __version__ as VERSION
-from .agent import AbstractAgent
 from .config import (
     agent_etcd_config_iv,
     agent_local_config_iv,
@@ -69,6 +75,9 @@ from .monitor import AgentErrorPluginContext, AgentStatsPluginContext
 from .types import AgentBackend, LifecycleEvent, VolumeInfo
 from .utils import get_subnet_ip
 
+if TYPE_CHECKING:
+    from .agent import AbstractAgent
+
 log = BraceStyleAdapter(logging.getLogger("ai.backend.agent.server"))
 
 deeplearning_image_keys = {
@@ -198,6 +207,7 @@ class AgentRPCServer(aobject):
 
         await self.read_agent_config()
         await self.read_agent_config_container()
+        await self.init_background_task_manager()
 
         self.stats_monitor = AgentStatsPluginContext(self.etcd, self.local_config)
         self.error_monitor = AgentErrorPluginContext(self.etcd, self.local_config)
@@ -283,6 +293,13 @@ class AgentRPCServer(aobject):
             self.local_config["container"][k] = v
             log.info("etcd: container-config: {}={}".format(k, v))
 
+    async def init_background_task_manager(self):
+        event_producer = await EventProducer.new(
+            cast(EtcdRedisConfig, self.local_config["redis"]),
+            db=4,  # Identical to manager's REDIS_STREAM_DB
+        )
+        self.local_config["background_task_manager"] = BackgroundTaskManager(event_producer)
+
     async def __aenter__(self) -> None:
         await self.rpc_server.__aenter__()
 
@@ -380,6 +397,8 @@ class AgentRPCServer(aobject):
                 "container_id": result["container_id"],
                 "resource_spec": result["resource_spec"],
                 "attached_devices": result["attached_devices"],
+                "agent_addr": result["agent_addr"],
+                "scaling_group": result["scaling_group"],
             }
             for result in results
         ]
@@ -390,7 +409,7 @@ class AgentRPCServer(aobject):
     async def destroy_kernel(
         self,
         kernel_id: str,
-        reason: str = None,
+        reason: Optional[KernelLifecycleEventReason] = None,
         suppress_events: bool = False,
     ):
         loop = asyncio.get_running_loop()
@@ -399,7 +418,7 @@ class AgentRPCServer(aobject):
         await self.agent.inject_container_lifecycle_event(
             KernelId(UUID(kernel_id)),
             LifecycleEvent.DESTROY,
-            reason or "user-requested",
+            reason or KernelLifecycleEventReason.USER_REQUESTED,
             done_future=done,
             suppress_events=suppress_events,
         )
@@ -500,6 +519,75 @@ class AgentRPCServer(aobject):
         log.info("rpc::start_service(k:{0}, app:{1})", kernel_id, service)
         return await self.agent.start_service(KernelId(UUID(kernel_id)), service, opts)
 
+    @rpc_function
+    @collect_error
+    async def get_commit_status(
+        self,
+        kernel_id,  # type: str
+        subdir,  # type: str
+    ):
+        # Only this function logs debug since web sends request at short intervals
+        log.debug("rpc::get_commit_status(k:{})", kernel_id)
+        status: CommitStatus = await self.agent.get_commit_status(
+            KernelId(UUID(kernel_id)),
+            subdir,
+        )
+        return {
+            "kernel": kernel_id,
+            "status": status.value,
+        }
+
+    @rpc_function
+    @collect_error
+    async def commit(
+        self,
+        kernel_id,  # type: str
+        subdir,  # type: str
+        filename,  # type: str
+    ):
+        log.info("rpc::commit(k:{})", kernel_id)
+        bgtask_mgr = self.local_config["background_task_manager"]
+        task_id = await bgtask_mgr.start(
+            self.agent.commit,
+            kernel_id=KernelId(UUID(kernel_id)),
+            subdir=subdir,
+            filename=filename,
+        )
+        return {
+            "bgtask_id": str(task_id),
+            "kernel": kernel_id,
+            "path": str(Path(subdir, filename)),
+        }
+
+    @rpc_function
+    @collect_error
+    async def get_local_config(self) -> Mapping[str, Any]:
+        agent_config: Mapping[str, Any] = self.local_config["agent"]
+        report_path: Path | None = agent_config.get("abuse-report-path")
+        return {
+            "agent": {
+                "abuse-report-path": str(report_path) if report_path is not None else "",
+            },
+            "watcher": self.local_config["watcher"],
+        }
+
+    @rpc_function
+    @collect_error
+    async def get_abusing_report(
+        self,
+        kernel_id,  # type: str
+    ) -> Mapping[str, str] | None:
+        if (abuse_path := self.local_config["agent"].get("abuse-report-path")) is not None:
+            report_path = Path(abuse_path, f"report.{kernel_id}.json")
+            if report_path.is_file():
+
+                def _read_file():
+                    with open(report_path, "r") as file:
+                        return json.load(file)
+
+                return await self.loop.run_in_executor(None, _read_file)
+        return None
+
     @rpc_function
     @collect_error
     async def shutdown_service(
@@ -522,6 +610,12 @@ class AgentRPCServer(aobject):
         log.info("rpc::download_file(k:{0}, fn:{1})", kernel_id, filepath)
         return await self.agent.download_file(KernelId(UUID(kernel_id)), filepath)
 
+    @rpc_function
+    @collect_error
+    async def download_single(self, kernel_id: str, filepath: str):
+        log.info("rpc::download_single(k:{0}, fn:{1})", kernel_id, filepath)
+        return await self.agent.download_single(KernelId(UUID(kernel_id)), filepath)
+
     @rpc_function
     @collect_error
     async def list_files(self, kernel_id: str, path: str):
@@ -574,6 +668,18 @@ class AgentRPCServer(aobject):
                 log.exception("reset: destroying {0}", kernel_id)
         await asyncio.gather(*tasks)
 
+    @rpc_function
+    @collect_error
+    async def assign_port(self):
+        log.debug("rpc::assign_port()")
+        return self.agent.port_pool.pop()
+
+    @rpc_function
+    @collect_error
+    async def release_port(self, port_no: int):
+        log.debug("rpc::release_port(port_no:{})", port_no)
+        self.agent.port_pool.add(port_no)
+
 
 @aiotools.server
 async def server_main_logwrapper(
@@ -681,6 +787,7 @@ async def server_main(
         hook_task_factory=local_config["debug"]["enhanced-aiomonitor-task-info"],
     )
     monitor.prompt = "monitor (agent) >>> "
+    monitor.console_locals["local_config"] = local_config
     monitor.start()
 
     # Start RPC server.
@@ -691,6 +798,7 @@ async def server_main(
         skip_detect_manager=local_config["agent"]["skip-manager-detection"],
     )
     agent_instance = agent
+    monitor.console_locals["agent"] = agent
 
     # Run!
     try:
@@ -805,6 +913,8 @@ def main(
             cfg["debug"]["coredump"]["core_path"] = Path(core_pattern).parent
 
         cfg["agent"]["pid-file"].write_text(str(os.getpid()))
+        image_commit_path = cfg["agent"]["image-commit-path"]
+        image_commit_path.mkdir(parents=True, exist_ok=True)
         ipc_base_path = cfg["agent"]["ipc-base-path"]
         log_sockpath = ipc_base_path / f"agent-logger-{os.getpid()}.sock"
         log_sockpath.parent.mkdir(parents=True, exist_ok=True)
